import {createMultipleUsers, createUser} from '../support/helpers';
import {randomUserData, type UserData} from '../support/rand';

const submitUserSearch = (user?: UserData): void => {
  cy.url().should('eq', `${Cypress.config().baseUrl}/`);
  cy.get('input[name="schoolSelection"]')
    .closest('.uFormElement')
    .within(() => {
      cy.get('.uButton')
        .click();
      cy.get('.uComboBox__list__option')
        .should('have.length.above', 1)
        .eq(1)
        .click();
    });
  cy.get('[name="quickSearch"]')
    .clear();
  if (user?.username) {
    cy.get('[name="quickSearch"]')
      .type(user.username);
  }
  cy.get('.searchForm__buttons button[type="submit"]')
    .should('have.length.above', 0)
    .click();
  cy.get('.uStandbyWrapper')
    .should('be.visible');
  cy.get('.uStandbyWrapper')
    .should('have.length', 0);
};

const _findUserRowIfExists = (user: UserData): Cypress.Chainable<JQuery<HTMLElement>> => cy.get('.uGrid').then($grid => {
  const $node = $grid.find(`:contains(${user.username})`);
  if ($node.length) {
    return $node.closest('tr');
  }
  return $node;
}).then($node => {
  if ($node.length > 0) {
    expect($node.find('td:nth-child(2)')).to.contain.text(user.firstname);
    expect($node.find('td:nth-child(3)')).to.contain.text(user.lastname);
    expect($node.find('td:nth-child(4)')).to.contain.text(user.username);
  }
  return $node;
});

const _setPage = (page: number): Cypress.Chainable<boolean> => cy.get('.uGrid').then($grid => {
  if (page < 1) {
    return cy.wrap(false).then(() => false);
  }
  const $select = $grid.find('.uTable__pagination select[name="page"]');
  if ($select.length === 0) {
    const result = page === 1;
    return cy.wrap(result).then(() => result);
  }
  const $lastOption = $select.find('option:last-child');
  const lastPage = $lastOption.length ? parseInt($lastOption.val() as string, 10) : 1;
  if (page > lastPage) {
    return cy.wrap(false).then(() => false);
  }
  const newValue = String(page);
  if ($select.val() === newValue) {
    return cy.wrap(true).then(() => true);
  }
  if ($select.val() === null) {
    // The select is disabled. This happens if there is no data in the table.
    //  We will return true iff page #1 was requested (as that is the only
    //  available option in such a case)
    return cy.wrap(newValue === '1').then(result => result);
  }
  cy.wrap($select).select(String(page));
  // The next line is flaky: we are waiting a tiny bit to make sure the table
  //  content will be re-rendered before moving on. It would be much nicer to
  //  have a [data-page="X"] attribute on the table DOM to avoid having to
  //  wait for the table to reload like this.
  // eslint-disable-next-line cypress/no-unnecessary-waiting
  return cy.wait(100).then(() => true);
});

const _determineCurrentPage = (): Cypress.Chainable<number> => cy.get('.uGrid').then($grid => {
  const $select = $grid.find('.uTable__pagination select[name="page"]');
  if ($select.length === 0) {
    return 1;
  }
  return parseInt($select.val() as string, 10);
});

const _determineLastPage = (): Cypress.Chainable<number> => cy.get('.uGrid').then($grid => {
  const $option = $grid.find('.uTable__pagination select[name="page"] option:last-child');
  if ($option.length === 0) {
    return 1;
  }
  return parseInt($option.val() as string, 10);
});

const _resetPaging = (position: 'first' | 'last' = 'first'): Cypress.Chainable => {
  if (position === 'first') {
    return _setPage(1);
  }
  return _determineLastPage().then(page => _setPage(page));
};

const _adjustPage = (diff: number): Cypress.Chainable<boolean> => _determineCurrentPage().then(page => _setPage(page + diff));

const _findUserRowByPaging = (user: UserData, direction: 1 | -1): Cypress.Chainable<JQuery<HTMLElement>> => _findUserRowIfExists(user).then($node => {
  if ($node.length) {
    return cy.wrap($node);
  }
  return _adjustPage(direction).then(foundNextPage => {
    if (foundNextPage) {
      return _findUserRowByPaging(user, direction);
    }
    return _findUserRowIfExists(user);
  });
});

const _findUserRowByPagingForward = (user: UserData): Cypress.Chainable<JQuery<HTMLElement>> => _findUserRowByPaging(user, 1);

const _findUserRowByPagingBackward = (user: UserData): Cypress.Chainable<JQuery<HTMLElement>> => _findUserRowByPaging(user, -1);

const findUserRow = (user: UserData, options?: {assertExists?: boolean; searchForward?: boolean}): Cypress.Chainable<JQuery<HTMLElement>> => _resetPaging(options?.searchForward ? 'first' : 'last').then(success => {
  expect(success).to.be.equal(true);
  if (options?.searchForward) {
    return _findUserRowByPagingForward(user);
  }
  return _findUserRowByPagingBackward(user);
}).then($row => {
  if (options?.assertExists) {
    expect($row).to.have.length.above(0, `Finding user "${user.username}"`);
    $row[0].scrollIntoView();
  }
  return $row;
});

describe('User handling tests', () => {

  it('creates and deletes users correctly', () => {
    cy.login();
    const userData = createUser();

    cy.get('.uGrid__tableWrapper tbody tr')
      .should('have.length', 0);
    cy.get('.uGrid__header .uButton')
      .eq(2)
      .should('have.attr', 'disabled');
    submitUserSearch(userData);
    const tr = findUserRow(userData, {assertExists: true});
    tr.find('td:first-child input').click();
    cy.get('.uGrid__header .uButton')
      .eq(2)
      .should('not.have.attr', 'disabled');
    cy.get('.uGrid__header .uButton')
      .eq(2)
      .click();
    cy.get('.uConfirmDialog .uButton--primary')
      .click();
    cy.get('.uGrid__tableWrapper tbody tr')
      .should('have.length', 0);
    submitUserSearch(userData);
    cy.get('.uGrid__tableWrapper tbody tr')
      .should('have.length', 0);
    cy.logout();
  });

  it('deletes multiple users correctly', () => {
    cy.login();
    const users = createMultipleUsers(2);
    submitUserSearch();
    findUserRow(users[0], {assertExists: true}).then($row => {
      cy.wrap($row).find('td:first-child input').click();
    });
    findUserRow(users[1], {assertExists: true}).then($row => {
      cy.wrap($row).find('td:first-child input').click();
    });
    cy.get('.uGrid__header .uButton')
      .eq(2)
      .should('not.have.attr', 'disabled');
    cy.get('.uGrid__header .uButton')
      .eq(2)
      .click();
    cy.get('.uConfirmDialog .uButton--primary')
      .click();
    submitUserSearch(users[0]);
    findUserRow(users[0]).should('have.length', 0);
    submitUserSearch(users[1]);
    findUserRow(users[1]).should('have.length', 0);
    cy.logout();
  });

  it('edits user correctly', () => {
    cy.login();
    const originalUserData = randomUserData();
    const editedUserData = randomUserData();
    createUser(originalUserData);
    submitUserSearch(originalUserData);
    cy.get('.uGrid__header .uButton')
      .eq(1)
      .should('have.attr', 'disabled');
    findUserRow(originalUserData, {assertExists: true})
      .find('td:first-child input')
      .click();
    cy.get('.uGrid__header .uButton')
      .eq(1)
      .within($button => {
        cy.wrap($button).should('not.have.attr', 'disabled');
        cy.wrap($button).click();
      });
    cy.url().should('eq', `${Cypress.config().baseUrl}/edit/${originalUserData.username}`);
    cy.get('input[name="firstname"]')
      .clear({scrollBehavior: 'center'})
      .type(editedUserData.firstname, {scrollBehavior: 'center'});
    cy.get('input[name="lastname"]')
      .clear({scrollBehavior: 'center'})
      .type(editedUserData.lastname, {scrollBehavior: 'center'});
    cy.get('input[name="name"]')
      .clear({scrollBehavior: 'center'})
      .type(editedUserData.username, {scrollBehavior: 'center'});
    const url = new URL(Cypress.config().baseUrl as string);
    cy.intercept({
      method: 'PATCH',
      hostname: url.hostname,
    }).as('patch');
    cy.get('.listView__header__buttons [type="button"]')
      .eq(0)
      .click();
    cy.wait('@patch');
    findUserRow(originalUserData).then($row => {
      cy.wrap($row).should('have.length', 0);
    });
    findUserRow(editedUserData, {assertExists: true});
    cy.logout();
  });

});
