/// <reference types="cypress" />

// ***********************************************
// This example commands.ts shows you how to
// create various custom commands and overwrite
// existing commands.
//
// For more comprehensive examples of custom
// commands please read more here:
// https://on.cypress.io/custom-commands
// ***********************************************
//
//
// -- This is a parent command --
// Cypress.Commands.add('login', (email, password) => { ... })
//
//
// -- This is a child command --
// Cypress.Commands.add('drag', { prevSubject: 'element'}, (subject, options) => { ... })
//
//
// -- This is a dual command --
// Cypress.Commands.add('dismiss', { prevSubject: 'optional'}, (subject, options) => { ... })
//
//
// -- This will overwrite an existing command --
// Cypress.Commands.overwrite('visit', (originalFn, url, options) => { ... })
//
// declare global {
//   namespace Cypress {
//     interface Chainable {
//       login(email: string, password: string): Chainable<void>
//       drag(subject: string, options?: Partial<TypeOptions>): Chainable<Element>
//       dismiss(subject: string, options?: Partial<TypeOptions>): Chainable<Element>
//       visit(originalFn: CommandOriginalFn, url: string, options: Partial<VisitOptions>): Chainable<Element>
//     }
//   }
// }

// Disabling some eslint rules:
// - camelcase: we don't control the naming of query-string parameters
// - no-unused-expressions: there are some assertions in here, which are "unused" expressions by definition
// - @typescript-eslint/no-unused-expressions: see above
/* eslint-disable camelcase, no-unused-expressions, @typescript-eslint/no-unused-expressions */

import sha256 from 'fast-sha256';

const uuid4 = (): string => [
  Math.random().toString(16).substring(2, 10),
  '4' + Math.random().toString(16).substring(2, 5), // eslint-disable-line prefer-template
  Math.random().toString(16).substring(2, 6),
  Math.random().toString(16).substring(2, 6),
  Math.random().toString(16).substring(2, 14),
].join('-');

const encodeChallenge = (codeVerifier: string): string => {
  const messageBuffer = new TextEncoder().encode(codeVerifier);
  return btoa(String.fromCharCode.apply(null, (sha256(messageBuffer) as unknown) as number[]))
    .replace(/\+/g, '-')
    .replace(/\//g, '_')
    .replace(/=+$/, '');
};

Cypress.Commands.add('login', (): void => {
  const root = Cypress.env('SSO_URI');
  const username = Cypress.env('SSO_USERNAME');
  const password = Cypress.env('SSO_PASSWORD');
  const realm = 'ucs';
  const state = uuid4();
  const codeVerifier = uuid4() + uuid4() + uuid4();
  cy.request({
    url: `${root}/realms/${realm}/protocol/openid-connect/auth`,
    qs: {
      client_id: 'ucs-school-ui-users',
      redirect_uri: 'http://localhost:4173/univention/ui/ucsschool-users/',
      scope: 'openid',
      state,
      nonce: uuid4(),
      response_type: 'code',
      response_mode: 'fragment',
      code_challenge: encodeChallenge(codeVerifier),
      code_challenge_method: 'S256',
    },
    followRedirect: false,
  }).then(response => {
    if (response.redirectedToUrl !== undefined) {
      // Already logged in. We could be strict here, and assume that this is an
      //  error, but we will choose to assume that an earlier test just failed,
      //  and the logout action just wasn't executed.
      return;
    }
    const html = document.createElement('html');
    html.innerHTML = response.body;

    const forms = html.getElementsByTagName('form');
    const isAuthorized = forms.length === 0;
    expect(isAuthorized).to.be.false;

    const form = forms[0];
    const requestBody = {
      username,
      password,
    };
    const hiddenInputs = form.querySelectorAll('input[type="hidden"]');
    for (const input of hiddenInputs) {
      const name = input.getAttribute('name');
      const value = input.getAttribute('value');
      if (name) {
        requestBody[name] = value;
      }
    }

    return cy.request({
      form: true,
      method: 'POST',
      url: form.action,
      followRedirect: false,
      body: requestBody,
    });
  }).then(response => {
    expect(response.redirectedToUrl).to.not.be.undefined;
    const codeMatch = /[?&#]code=(?<value>[^&]+)/.exec(response.redirectedToUrl as string);
    expect(codeMatch).to.not.be.null;
    const code: string = (codeMatch as RegExpExecArray)[1];
    const stateMatch = /[?&#]state=(?<value>[^&]+)/.exec(response.redirectedToUrl as string);
    expect(stateMatch).to.not.be.null;
    const stateResponse: string = (stateMatch as RegExpExecArray)[1];
    expect(stateResponse).to.equal(state);
    return cy.request({
      url: `${root}/realms/${realm}/protocol/openid-connect/token`,
      method: 'POST',
      body: {
        state,
        code,
        code_verifier: codeVerifier,
        grant_type: 'authorization_code',
        client_id: 'ucs-school-ui-users',
        redirect_uri: 'http://localhost:4173/univention/ui/ucsschool-users/',
      },
      form: true,
      followRedirect: false,
    });
  }).then(response => {
    cy.wrap(response.body.id_token, {log: false}).as('_ssoIdToken');
  });
});

Cypress.Commands.add('logout', (): void => {
  const root = Cypress.env('SSO_URI');
  const realm = 'ucs';
  cy.get('@_ssoIdToken', {log: false}).then(idToken => cy.request({
    url: `${root}/realms/${realm}/protocol/openid-connect/logout`,
    qs: {
      client_id: 'ucs-school-ui-users',
      id_token_hint: idToken,
    },
    followRedirect: false,
  })).then(response => {
    expect(response.redirectedToUrl).to.be.undefined;
    const html = document.createElement('html');
    html.innerHTML = response.body;
    const forms = document.getElementsByTagName('form');
    expect(forms.length).to.equal(0);
  });
});
